/*
 * Copyright 2017-2024 original authors
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * https://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package example.micronaut.chess

import example.micronaut.chess.dto.Player
import example.micronaut.chess.dto.GameDTO
import example.micronaut.chess.dto.GameStateDTO
import groovy.transform.CompileStatic
import io.micronaut.http.annotation.Controller
import io.micronaut.http.annotation.PathVariable
import io.micronaut.http.annotation.Post
import io.micronaut.http.annotation.Status
import io.micronaut.scheduling.TaskExecutors
import io.micronaut.scheduling.annotation.ExecuteOn
import reactor.core.publisher.Mono

import static io.micronaut.http.HttpStatus.CREATED
import static io.micronaut.http.HttpStatus.NO_CONTENT
import static io.micronaut.http.MediaType.APPLICATION_FORM_URLENCODED
import static io.micronaut.http.MediaType.TEXT_PLAIN

@Controller('/game') // <1>
@ExecuteOn(TaskExecutors.BLOCKING) // <2>
@CompileStatic
class GameController {

    private final GameReporter gameReporter

    GameController(GameReporter gameReporter) {
        this.gameReporter = gameReporter
    }

    @Post(value = '/start', // <3>
          consumes = APPLICATION_FORM_URLENCODED, // <4>
          produces = TEXT_PLAIN) // <5>
    @Status(CREATED) // <6>
    Mono<String> start(String b, // <7>
                       String w) {
        GameDTO game = new GameDTO(UUID.randomUUID().toString(), b, w) // <8>
        gameReporter.game(game.id, game).map(gameDTO -> game.id) // <9>
    }

    @Post(value = '/move/{gameId}', // <10>
          consumes = APPLICATION_FORM_URLENCODED) // <11>
    @Status(CREATED) // <12>
    void move(@PathVariable String gameId,
              Player player,
              String move,
              String fen,
              String pgn) {
        GameStateDTO gameState = new GameStateDTO(UUID.randomUUID().toString(),
                gameId, player, move, fen, pgn)
        gameReporter.gameState(gameId, gameState).subscribe() // <13>
    }

    @Post('/checkmate/{gameId}/{player}') // <14>
    @Status(NO_CONTENT) // <15>
    void checkmate(@PathVariable String gameId,
                   @PathVariable Player player) {
        GameDTO game = new GameDTO(gameId, false, player)
        gameReporter.game(gameId, game).subscribe() // <16>
    }

    @Post('/draw/{gameId}') // <17>
    @Status(NO_CONTENT) // <18>
    void draw(@PathVariable String gameId) {
        GameDTO game = new GameDTO(gameId, true, null)
        gameReporter.game(gameId, game).subscribe() // <19>
    }
}
